# pretty-printer for BigPtrArray

from openoffice.util import register

import gdb


class BigPtrArrayPrinter:
    """Prints a BigPtrArray."""

    def __init__(self, value):
        self.value = value

    def to_string(self):
        typename = self.value.type.tag
        length = int(self.value['nSize'])
        if length > 0:
            return "%s of length %d" % (typename, length)
        else:
            return "empty %s" % typename

    def children(self):
        return self._iterator(self.value)

    def display_hint(self):
        return 'array'


    class _iterator:

        def __init__(self, array):
            self.blocks = array['ppInf']
            self.count = array['nSize']
            self.pos = 0
            self.block_count = array['nBlock']
            self.block_pos = 0
            self.block = None
            self._next_block(False)
            self._check_invariant()

        def __iter__(self):
            return self

        def next(self):
            if self.pos == self.count:
                raise StopIteration()

            name = str(self.pos)
            value = self.block['pData'][self.pos - self.block['nStart']]
            if self.pos == self.block['nEnd']:
                self._next_block()
            self.pos += 1

            self._check_invariant()
            return (name, value)

        def _next_block(self, advance = True):
            if advance:
                self.block_pos += 1

            if self.block_pos == self.block_count:
                return

            pblock = self.blocks[self.block_pos]
            assert pblock
            block = pblock.dereference()
            start = block['nStart']
            end = block['nEnd']
            assert end - start + 1 == block['nElem']
            if self.block:
                assert start == self.block['nEnd'] + 1
                assert end <= self.count
            else:
                assert start == 0
            self.block = block

        def _check_invariant(self):
            assert self.pos <= self.count
            assert self.block_pos <= self.block_count
            if self.pos == 0 and self.pos < self.count:
                assert self.block != None


def register_pretty_printers(obj):
    register.register_for_pattern(obj, 'BigPtrArray',
            lambda val: BigPtrArrayPrinter(val))
