# pretty printers for old tools' containers

from openoffice.util import register

import gdb


class IteratorHelper:
    """Implements a container iterator useable for both 'linear'
        containers (like DynArray or List) and Tables.
    """

    def __init__(self, block, count, type = None):
        #assert type

        self.count = count
        self.type = type
        self.pos = 0
        self.block = None
        self.block_count = 0
        self.block_pos = 0
        if block:
            self._next_block(block)

        self._check_invariant()

    def __iter__(self):
        return self

    def next(self):
        if self.pos == self.count:
            raise StopIteration()

        if self.block_pos == self.block_count:
            self._next_block(self.block['pNext'])

        name = self.name()
        value = self.value()
        self.advance()

        self._check_invariant()
        return (name, value)

    def _next_block(self, block):
        assert block

        self.block = block.dereference()
        self.block_pos = 0
        self.block_count = block['nCount']

        assert self.block_count <= block['nSize']
        assert self.block_count + self.pos <= self.count

    def _check_invariant(self):
        assert self.count >= 0
        assert self.pos >= 0
        assert self.pos <= self.count
        assert self.block_count >= 0
        if self.pos < self.count:
            assert self.block_count > 0
            assert self.block != None
        assert self.block_count <= self.count
        assert self.block_pos >= 0
        assert self.block_pos <= self.block_count


class NoItemType (Exception):
    pass


class ContainerHelper:
    """Provides support for specialized container printers."""

    def to_string(self):
        size = self.value['nCount']
        if size > 0:
            return "%s of length %s" % (self.typename, size)
        elif size == 0:
            return "empty %s" % self.typename
        else:
            return "invalid %s" % self.typename

    def children(self):
        count = self.value.cast(gdb.lookup_type('Container'))['nCount']
        return self.iterator(self.value['pFirstBlock'], count)


class LinearIterator (IteratorHelper):
    """Is an iterator for a 'linear' container."""

    def __init__(self, block, count, type = None):
        IteratorHelper.__init__(self, block, count, type)

    def name(self):
        return str(self.pos)

    def value(self):
        nodes = self.block['pNodes']#.cast(self.type.pointer())
        return nodes[self.block_pos]

    def advance(self):
        self.pos += 1
        self.block_pos += 1


class LinearContainerPrinter (ContainerHelper):
    """Prints a 'linear' container, like DynArray or List."""

    def __init__(self, typename, value):
        self.value = value
        self.typename = typename
        self.iterator = LinearIterator

    def display_hint(self):
        return 'array'


class TableIterator (IteratorHelper):
    """Is an iterator for a Table."""

    def __init__(self, block, count, type = None):
        IteratorHelper.__init__(self, block, count, type)
        # ULONG doesn't work on 64-bit for some reason (gdb says it has
        # size 4 and it's not a typedef to sal_uIntPtr)
        self._key_type = gdb.lookup_type('sal_uIntPtr')
        self.is_key = True

    def name(self):
        return ''

    def value(self):
        nodes = self.block['pNodes']#.cast(self.type.pointer())
        value = nodes[self.block_pos]
        if self.is_key:
            value = str(value.cast(self._key_type))
        return value

    def advance(self):
        self.pos += 1
        self.block_pos += 1
        self.is_key = not self.is_key


class TablePrinter (ContainerHelper):
    """Prints a Table."""

    def __init__(self, value):
        self.value = value
        self.typename = 'Table'
        self.iterator = TableIterator

    def display_hint(self):
        return 'map'


def register_pretty_printers(obj):
    default = register.default_filter()
    def addpp(name, printer):
        register.register(obj, printer, register.base_class_tester(name, default))

    addpp('Container', lambda val: LinearContainerPrinter('Container', val))
    addpp('DynArray', lambda val: LinearContainerPrinter('DynArray', val))
    addpp('List', lambda val: LinearContainerPrinter('List', val))
    addpp('Queue', lambda val: LinearContainerPrinter('Queue', val))
    addpp('Stack', lambda val: LinearContainerPrinter('Stack', val))
    addpp('Table', lambda val: TablePrinter(val))
