from openoffice.util import register
from openoffice.util.string import Rtl_String, Rtl_UString, RtlOString, RtlOUString, StringEncoder, UnicodeEncoder

import gdb


class SalUnicodeStringPrinter:
    """Prints a sal_Unicode*"""

    def __init__(self, value):
        self.value = value

    def to_string(self):
        len = self._strlen(self.value)
        chartype = self.value.type.target()
        nchartype = gdb.lookup_type('sal_Char')
        encoder = UnicodeEncoder(chartype, nchartype)
        return encoder.encode(self.value, len)

    def display_hint(self):
        return 'string'

    def _strlen(self, string):
        len = 0
        while string[len] != 0:
            len += 1
        return len


class Rtl_StringPrinter:
    """Prints an rtl_String or rtl_uString"""

    def __init__(self, encoding, is_unicode, value):
        self.value = value
        self.encoding = encoding
        if is_unicode:
            self.ooo_string = Rtl_UString
        else:
            self.ooo_string = Rtl_String

    def to_string(self):
        string = self.__string()
        if string:
            return string
        else:
            return "uninitialized string"

    def display_hint(self):
        if self.__string():
            return 'string'
        else:
            return ''

    def __string(self):
        return self.ooo_string(self.value, self.encoding).string()


class RtlStringPrinter:
    """Prints an rtl::OString or rtl::OUString"""

    def __init__(self, encoding, is_unicode, value):
        self.value = value
        self.encoding = encoding
        if is_unicode:
            self.ooo_string = RtlOUString
        else:
            self.ooo_string = RtlOString

    def to_string(self):
        string = self.__string()
        if string:
            return string
        else:
            return "uninitialized string"

    def display_hint(self):
        if self.__string():
            return 'string'
        else:
            return ''

    def __string(self):
        return self.ooo_string(self.value, self.encoding).string()


class RtlStringBufferPrinter:
    """Prints an rtl::OStringBuffer or rtl::OUStringBuffer."""

    def __init__(self, encoding, is_unicode, value):
        self.value = value
        self.encoding = encoding
        self.is_unicode = is_unicode

    def to_string(self):
        typename = self.value.type.tag
        data = self.value['pData']
        capacity = self.value['nCapacity']
        if data:
            string = data.dereference()
            return self._make_string()
        else:
            return "unintialized %s" % typename

    def _make_string(self):
        if self.is_unicode:
            string = Rtl_UString(self.value, self.encoding)
        else:
            string = Rtl_String(self.value, self.encoding)
        return string.string()

    def display_hint(self):
        if self.value['pData']:
            return 'string'
        else:
            return ''


class ToolsStringPrinter:
    """Prints a ByteString or UniString"""

    def __init__(self, encoding, is_unicode, value):
        self.value = value
        self.encoding = encoding
        self.is_unicode = is_unicode

    def to_string(self):
        if self.is_valid():
            return self.make_string()
        else:
            return "uninitialized string"

    def display_hint(self):
        if self.is_valid():
            return 'string'
        else:
            return ''

    def is_valid(self):
        data = self.value['mpData']
        # mnRefCount is not a good indicator: it seems there could be
        # cases where it is negative (-7FFFFFED)
        return data #and data.dereference()['mnRefCount'] > 0
        return True

    def make_string(self):
        assert self.value['mpData']
        data = self.value['mpData'].dereference()
        chartype = data['maStr'].type.target()
        if self.is_unicode:
            nchartype = gdb.lookup_type('sal_Char')
            encoder = UnicodeEncoder(chartype, nchartype, self.encoding)
            return encoder.encode(data['maStr'], data['mnLen'])
        else:
            ptrtype = chartype.pointer()
            encoder = StringEncoder(self.encoding)
            return encoder.encode(data['maStr'].cast(ptrtype))


def register_pretty_printers(obj):
    def addpp(pat, pr):
        register.register_for_pattern(obj, pat, pr)
    addpp('^_rtl_String$', lambda val: Rtl_StringPrinter('@target-charset', False, val))
    addpp('^_rtl_uString$', lambda val: Rtl_StringPrinter('utf-16', True, val))
    addpp('^rtl::OString$', lambda val: RtlStringPrinter('@target-charset', False, val))
    addpp('^rtl::OUString$', lambda val: RtlStringPrinter('utf-16', True, val))
    addpp('^rtl::OStringBuffer$', lambda val: RtlStringBufferPrinter('@target-charset', False, val))
    addpp('^rtl::OUStringBuffer$', lambda val: RtlStringBufferPrinter('utf-16', True, val))
    addpp('^ByteString$', lambda val: ToolsStringPrinter('@target-charset', False, val))
    addpp('^String$', lambda val: ToolsStringPrinter('utf-16', True, val))

    def target_test(type):
        if type.code == gdb.TYPE_CODE_REF:
            return type
        return None
    def name_test(type):
        if str(type) == 'sal_Unicode':
            return type
        return None
    name_tester = register.make_filter(name_test)
    target_tester = register.make_filter(target_test)
    register.register(obj, lambda val: SalUnicodeStringPrinter(val),
            name_tester(register.target_mapper(target_tester(
                register.unqualified_mapper()))))
