# Pretty printers for SvArray types.

from openoffice.util import register

import gdb


class SvArrayPrinter:
    """Prints macro-declared arrays from svl module."""

    def __init__(self, value):
        self.value = value

    def to_string(self):
        typename = self.value.type.tag
        if int(self.value['nA']):
            return "%s of length %d" % (typename, int(self.value['nA']))
        else:
            return "empty " + typename

    def children(self):
        return self._iterator(self.value['pData'], int(self.value['nA']))

    def display_hint(self):
        return 'array'

    class _iterator:

        def __init__(self, data, count):
            self.data = data
            self.count = count
            self.pos = 0
            self._check_invariant()

        def __iter__(self):
            return self

        def next(self):
            if self.pos == self.count:
                raise StopIteration()

            pos = self.pos
            elem = self.data[pos]
            self.pos = self.pos + 1

            self._check_invariant()
            return (str(pos), elem)

        def _check_invariant(self):
            assert self.count >= 0
            if self.count > 0:
                assert self.data
            assert self.pos >= 0
            assert self.pos <= self.count


def svarray_test(type):
    if not type.tag:
        return None

    ushort = gdb.lookup_type('USHORT')
    ret = True
    for field in type.fields():
        if field.name == 'pData':
            ret = field.type.code == gdb.TYPE_CODE_PTR
        elif field.name == 'nFree':
            ret = str(field.type) == ushort
        elif field.name == 'nA':
            ret = str(field.type) == ushort
        else:
            ret = False
        if not ret:
            return None

    try:
        gdb.lookup_type('FnForEach_' + type.tag)
    except RuntimeError:
        return None

    return type

svarray_tester = register.make_filter(svarray_test)


def name_tester(name, filter = None):
    def test(type):
        if str(type) == name:
            return type
        return None
    return register.FilterHelper(test, filter)


def register_pretty_printers(obj):
    # Register printer for predeclared arrays
    predeclared_arrays = ['SvBools', 'SvBytes', 'SvULongs','SvULongsSort',
            'SvUShorts', 'SvUShortsSort', 'SvLongs', 'SvLongsSort',
            'SvShorts', 'SvStrings', 'SvStringsDtor', 'SvStringsSort',
            'SvStringsSortDtor', 'SvStringsISort', 'SvStringsISortDtor',
            'SvByteStrings', 'SvByteStringsDtor', 'SvByteStringsSort',
            'SvByteStringsSortDtor', 'SvByteStringsISort',
            'SvByteStringsISortDtor', 'SvXub_StrLens', 'SvXub_StrLensSort']
    for array in predeclared_arrays:
        register.register(obj, lambda val: SvArrayPrinter(val),
                name_tester(array,
                    register.unqualified_mapper(register.target_mapper())))

    # Register printer for pointer arrays
    register.register(obj, lambda val: SvArrayPrinter(val),
            register.base_class_tester('SvPtrarr',
                register.unqualified_mapper(register.target_mapper())))
    register.register(obj, lambda val: SvArrayPrinter(val),
            register.base_class_tester('SvPtrarrPlain',
                register.unqualified_mapper(register.target_mapper())))

    # Register printer for var and obj arrays
    register.register(obj, lambda val: SvArrayPrinter(val),
            svarray_tester(register.default_filter()))
