# pretty printers for old tools' types

from openoffice.util import register

import gdb


class BigIntPrinter:
    """Prints a big integer"""

    def __init__(self, value):
        self.value = value

    def to_string(self):
        if self.value['bIsSet']:
            if self.value['bIsBig']:
                return self.bigint()
            else:
                return self.value['nVal']
        else:
            return "unset %s" % self.value.type.tag

    def bigint(self):
        len = self.value['nLen']
        digits = self.value['nNum']
        num = 0
        dsize = digits.dereference().type.sizeof
        # The least significant word is on index 0
        for i in reverse(range(0, len - 1)):
            num *= dsize
            num += digits[i]
        return num


class ColorPrinter:
    """Print color as rgb(r, g, b) or rgba(r, g, b, a)"""

    def __init__(self, value):
        self.value = value

    def to_string(self):
        color = self.value['mnColor']
        b = color & 0xff
        g = (color >> 8) & 0xff
        r = (color >> 16) & 0xff
        a = (color >> 24) & 0xff
        if a:
            return "rgba(%d, %d, %d, %d)" % (r, g, b, a)
        else:
            return "rgb(%d, %d, %d)" % (r, g, b)


class DateTimeImpl:

    def __init__(self, date, time):
        self.date = date
        self.time = time

    def __str__(self):
        result = ""
        if self.date:
            result += str(self.date)
        if self.date and self.time:
            result += 'T'
        if self.time:
            result += str(self.time)
        return result


def parse_datetime(value):
    return DateTimeImpl(parse_date(value), parse_time(value))


class DateTimePrinter:

    def __init__(self, value):
        self.value = value

    def to_string(self):
        return str(parse_datetime(self.value))


class DateImpl:

    def __init__(self, year, month, day):
        self.year = year
        self.month = month
        self.day = day

    def __str__(self):
        return "%d-%02d-%02d" % (self.year, self.month, self.day)


def parse_date(value):
    date = value['nDate']
    d = date % 100
    m = (date / 100) % 100
    y = date / 10000
    return DateImpl(y, m, d)


class DatePrinter:

    def __init__(self, value):
        self.value = value

    def to_string(self):
        return str(parse_date(self.value))


class FractionPrinter:

    def __init__(self, value):
        self.value = value

    def to_string(self):
        numerator = self.value['nNumerator']
        denominator = self.value['nDenominator']
        if denominator > 0:
            return "%d/%d" % (numerator, denominator)
        else:
            return "invalid Fraction"


class TimeImpl:

    def __init__(self, hour, minute, second, hundreth):
        self.hour = hour
        self.minute = minute
        self.second = second
        self.hundreth = hundreth

    def __str__(self):
        time = "%02d:%02d:%02d" % (self.hour, self.minute, self.second)
        if self.hundreth != 0:
            time += ".%02d" % self.hundreth
        return time


def parse_time(value):
    time = value['nTime']
    h = time / 1000000
    m = (time / 10000) % 100
    s = (time / 100) % 100
    f = time % 100
    return TimeImpl(h, m, s, f)


class TimePrinter:

    def __init__(self, value):
        self.value = value

    def to_string(self):
        return str(parse_time(self.value))


def register_pretty_printers(obj):
    def addpp(pat, pr):
        register.register_for_pattern(obj, pat, pr)
    addpp('^BigInt$', lambda val: BigIntPrinter(val))
    addpp('^Color$', lambda val: ColorPrinter(val))
    addpp('^Fraction$', lambda val: FractionPrinter(val))
    addpp('^DateTime$', lambda val: DateTimePrinter(val))
    addpp('^Date$', lambda val: DatePrinter(val))
    addpp('^Time$', lambda val: TimePrinter(val))
