from openoffice.util import register
from openoffice.util.uno import make_uno_type, uno_cast

import gdb


class UnoAnyPrinter:
    """Prints a UNO any."""

    def __init__(self, value):
        self.value = value
        self.typename = "UNO any"

    def to_string(self):
        if self.is_set():
            return ('%s %s' % (self.typename, self.make_string()))
        else:
            return "empty %s" % self.typename

    def is_set(self):
        return self.value['pData'] and self.value['pType']

    def make_string(self):
        ptr = self.value['pData']
        assert ptr
        type_desc = self.value['pType']
        assert type_desc
        type = make_uno_type(type_desc.dereference())
        assert type
        return str(uno_cast(type, ptr).dereference())


class RtlReferencePrinter:
    """Prints an rtl::Reference<>"""

    def __init__(self, value):
        self.value = value

    def to_string(self):
        typename = 'rtl::Reference'
        pointee = self.value['m_pBody']
        if pointee:
            val = pointee.dereference()
            return '%s to %s' % (typename, str(val))
        else:
            return "empty %s" % typename


class UnoReferencePrinter:
    """Prints a reference to a UNO interface."""

    def __init__(self, value):
        self.value = value
        self.typename = 'reference to UNO interface'

    def to_string(self):
        iface = self.value['_pInterface']
        if iface:
            impl = iface.cast(self._itype()).dereference()
            return '%s to %s' % (self.typename, str(impl))
        else:
            return "empty %s" % self.typename

    def _itype(self):
        return self.value.type.template_argument(0).pointer()


class UnoSequencePrinter:
    """Prints a UNO Sequence."""

    class iterator:
        """Sequence iterator"""

        def __init__(self, first, size):
            self.item = first
            self.size = size
            self.count = 0

        def __iter__(self):
            return self

        def next(self):
            if self.count == self.size:
                raise StopIteration
            count = self.count
            self.count = self.count + 1
            elem = self.item.dereference()
            self.item = self.item + 1
            return ('[%d]' % count, elem)


    def __init__(self, value):
        self.value = value
        self.typename = "UNO sequence"

    def to_string(self):
        pimpl = self.value['_pSequence']
        if pimpl:
            impl = pimpl.dereference()
            elems = impl['nElements']
            if elems == 0:
                return "empty %s" % self.typename
            else:
                return "%s of length %s" % (self.typename, elems)
        else:
            return "uninitialized %s" % self.typename

    def children(self):
        pimpl = self.value['_pSequence']
        if pimpl:
            impl = pimpl.dereference()
            elemtype = self.value.type.template_argument(0)
            elements = impl['elements'].cast(elemtype.pointer())
            return self.iterator(elements, int(impl['nElements']))
        else:
            # TODO is that the best thing to do here?
            return None

    def display_hint(self):
        if self.value['_pSequence']:
            return 'array'
        else:
            return ''


class UnoTypePrinter:
    """Prints a UNO Type."""

    def __init__(self, value):
        self.value = value
        self.typename = "UNO type"

    def to_string(self):
        uno = make_uno_type(self.value)
        if uno:
            return "%s %s" % (self.typename, uno.tag)
        else:
            return "unitialized %s" % self.typename


def register_pretty_printers(obj):
    def addpp(pat, pr):
        register.register_for_pattern(obj, pat, pr)
    addpp('^_uno_Any$', lambda val: UnoAnyPrinter(val))
    addpp('^com::sun::star::uno::Any$', lambda val: UnoAnyPrinter(val))
    addpp('^rtl::Reference<.*>$', lambda val: RtlReferencePrinter(val))
    addpp('^com::sun::star::uno::Reference<.*>$', lambda val: UnoReferencePrinter(val))
    addpp('^com::sun::star::uno::Sequence<.*>$', lambda val: UnoSequencePrinter(val))
    addpp('^com::sun::star::uno::Type$', lambda val: UnoTypePrinter(val))
