import gdb
import re


class PrettyPrinterLookup:
    """Is an association of a list of pretty printers with an object file."""

    def __init__(self, obj):
        self.list = []
        self.obj = obj

    def __eq__(self, other):
        # There is just one pretty printers' dictionary for an (gdb)
        # object
        return self.obj == other.obj

    def __hash__(self):
        return self.obj.__hash__()

    def __getitem__(self, idx):
        return self.list.__getitem__(idx)

    def __setitem__(self, idx, value):
        return self.list.__setitem__(idx, value)

    def __iter__(self):
        return self.list.__iter__()

    def __contains__(self, idx):
        return self.list.__contains__(idx)

    def append(self, item):
        self.list.append(item)


class FilterHelper:
    """Is a helper for creating filters. See lookup_function for details."""

    def __init__(self, filter, chained = None):
        self.filter = filter
        self.chained = chained

    def __call__(self, type):
        if type and self.chained:
            type = self.chained(type)
        if type:
            type = self.filter(type)
        return type


def make_filter(filter):
    """Creates a new filter functor from function filter.
        Examples:
        unqualified_mapper = make_filter(lambda t: t.unqualified())
    """
    def make(chained = None):
        return FilterHelper(filter, chained)
    return make


def target_mapper(chained = None):
    """Removes all pointers/references."""
    def map(type):
        if type.code == gdb.TYPE_CODE_REF:
            return type.target()
        else:
            return type
    return FilterHelper(map, chained)

# Removes all qualifications.
unqualified_mapper = make_filter(lambda t: t.unqualified())
# Strips typedefs.
strip_typedefs_mapper = make_filter(lambda t: t.strip_typedefs())

# Is a default filter to be used.
default_filter = make_filter(
        strip_typedefs_mapper(unqualified_mapper(target_mapper())))

def regex_tester(pattern, filter = None):
    """Matches against regex."""
    def test(type):
        if type.tag:
            if pattern.search(type.tag):
                return type
        return None
    return FilterHelper(test, filter)


def base_class_tester(name, filter = None):
    def test(type):
        assert type
        if type.code == gdb.TYPE_CODE_STRUCT:
            for field in type.fields():
                if field.name == name:
                    return field.type
        return None
    return FilterHelper(test, filter)


def register(obj, cons, filter = default_filter()):
    """Registers a new pretty printer for object file.
        @obj the object file
        @cons the pretty printer
        @filter the filter to use for lookup
    """
    if obj == None:
        obj = gdb

    if not obj in pretty_printers_dict:
        lookup = PrettyPrinterLookup(gdb)
        pretty_printers_dict[obj] = lookup
        obj.pretty_printers.append(lambda val: lookup_function(val, lookup))

    pretty_printers_dict[obj].append((cons, filter))


def register_for_pattern(obj, pat, cons, filter = default_filter()):
    if isinstance(pat, str):
        pat = re.compile(pat)
    register(obj, cons, regex_tester(pat, filter))


def lookup_function(val, pps):
    """Look-up and return a pretty-printer that can print val.
        It uses so-called filter to determine if a pretty-printer is
        suitable for given value. There are two kinds of filters:
        * mapper -- gets a gdb.Type and produces (possibly the same) gdb.Type
        * tester -- gets a gdb.Type and produces either that type or None.
        It's possible to chain filters to implement more complicated
        ones; usually there would be several mappers followed by one
        tester. For any of them, if the input is None, it's just passed
        through.

        The lookup loops through registered pretty-printers, for each of
        them it performs the associated filter and if it doesn't return
        None, the pretty printer is used; otherwise the lookup continues
        by the next pretty printer.
    """

    for (cons, filter) in pps:
        if filter(val.type):
            return cons(val)

    # Cannot find a pretty printer
    return None


pretty_printers_dict = {}

# vim: set sts=4 sw=4 et:
