import gdb


class StringEncoder:

    def __init__(self, encoding):
        self.encoding = encoding

    def encode(self, value):
        # Look up the target encoding as late as possible.
        encoding = self.encoding
        if encoding[0] is '@':
            encoding = gdb.parameter(encoding[1:])
        return value.string(encoding)


class UnicodeEncoder:

    def __init__(self, srctype, tgttype, encoding = 'utf-16'):
        # TODO: throw out the types
        self.source_type = srctype
        self.target_type = tgttype
        self.encoding = encoding

    def encode(self, uniptr, unilen):
        result = u''
        for i in range(0, int(unilen)):
            result += unichr(uniptr[i])
        return result


class Rtl_StringBase:
    """Base class for rtl_?String representations"""

    def __init__(self, value, encoding):
        self.value = value
        self.encoding = encoding

    def type(self):
        return self.value.type

    def string(self):
        if self.is_valid():
            return self.make_string()
        else:
            return None

    def is_valid(self):
        return self.value['refCount'] > 0

    def __unicode__(self):
        return self.make_string()


class Rtl_String (Rtl_StringBase):
    """Represents an rtl_String"""

    def __init__(self, value, encoding = 'ascii'):
        Rtl_StringBase.__init__(self, value, encoding)

    def make_string(self):
        buf = self.value['buffer']
        chartype = buf.type.target()
        ptrtype = chartype.pointer()
        data = buf.cast(ptrtype)
        encoder = StringEncoder(self.encoding)
        return encoder.encode(data)


class Rtl_UString (Rtl_StringBase):
    """Represents an rtl_uString"""

    def __init__(self, value, encoding = 'utf-16'):
        Rtl_StringBase.__init__(self, value, encoding)

    def make_string(self):
        buf = self.value['buffer']
        chartype = buf.type.target()
        nchartype = gdb.lookup_type('sal_Char')
        encoder = UnicodeEncoder(chartype, nchartype, self.encoding)
        return encoder.encode(buf, self.value['length'])


class RtlOStringBase:
    """Base class for rtl::O?String representations"""

    def __init__(self, value, encoding, backend):
        self.value = value
        self.encoding = encoding
        self.backend = backend

    def type(self):
        return self.value.type

    def string(self):
        if self.is_valid():
            return self.make_string()
        else:
            return None

    def is_valid(self):
        return self.value['pData']

    def make_string(self):
        assert self.value['pData']
        data = self.value['pData'].dereference()
        return self.backend(data, self.encoding).string()

    def __unicode__(self):
        return self.make_string()


class RtlOString (RtlOStringBase):
    """Represents an rtl::OString"""

    def __init__(self, value, encoding = 'ascii'):
        RtlOStringBase.__init__(self, value, encoding, Rtl_String)


class RtlOUString (RtlOStringBase):
    """Represents an rtl::OUString"""

    def __init__(self, value, encoding = 'utf-16'):
        RtlOStringBase.__init__(self, value, encoding, Rtl_UString)


def make_string(value):
    """Creates a Python string from any oo.o string representation."""

    typename = value.type.strip_typedefs().tag
    if typename == '_rtl_String':
        string = Rtl_String(value)
    elif typename == '_rtl_uString':
        string = Rtl_UString(value)
    elif typename == 'rtl::OString':
        string = RtlOString(value)
    elif typename == 'rtl::OUString':
        string = RtlOUString(value)
    else:
        string = None

    if string:
        string = unicode(string)
    return string
