import gdb
from openoffice.util.string import make_string


class UnsupportedType (Exception):
    """Represents exception thrown when an unsupported UNO type (like
        array or union) is used."""

    def __init__(self, type):
        self.type = type


class UnknownType (Exception):
    """Represents exception thrown when an unknown UNO type is used."""

    def __init__(self, type):
        self.type = type


class TypeCode:
    """Represents type code (type class) of UNO type."""

    # type class of void
    VOID = 0
    # type class of char
    CHAR = 1
    # type class of boolean
    BOOLEAN = 2
    # type class of byte
    BYTE = 3
    # type class of short
    SHORT = 4
    # type class of unsigned short
    UNSIGNED_SHORT = 5
    # type class of long
    LONG = 6
    # type class of unsigned long
    UNSIGNED_LONG = 7
    # type class of hyper
    HYPER = 8
    # type class of unsigned hyper
    UNSIGNED_HYPER = 9
    # type class of float
    FLOAT = 10
    # type class of double
    DOUBLE = 11
    # type class of string
    STRING = 12
    # type class of type
    TYPE = 13
    # type class of any
    ANY = 14
    # type class of enum
    ENUM = 15
    # type class of typedef
    TYPEDEF = 16
    # type class of struct
    STRUCT = 17
    # type class of union (not implemented)
    UNION = 18
    # type class of exception
    EXCEPTION = 19
    # type class of sequence
    SEQUENCE = 20
    # type class of array (not implemented)
    ARRAY = 21
    # type class of interface
    INTERFACE = 22
    # type class of service (not implemented)
    SERVICE = 23
    # type class of module (not implemented)
    MODULE = 24
    # type class of interface method
    INTERFACE_METHOD = 25
    # type class of interface attribute
    INTERFACE_ATTRIBUTE = 26
    # type class of unknown type
    UNKNOWN = 27
    # type class of properties
    PROPERTY = 28
    # type class of constants
    CONSTANT = 29
    # type class of constants groups
    CONSTANTS = 30
    # type class of singletons
    SINGLETON = 31


class Type:
    """Describes a UNO type."""

    def __init__(self, code, tag):
        """Constructs a new Type.
            @param[in] code value of com::sun::star::uno::TypeClass
            @param[in] tag UNO name of the type
        """
        self.code = int(code)
        self.tag = str(tag)


    def to_gdb(self):
        """Gets gdb.Type for the type."""
        return gdb.lookup_type(self.typename)


def make_uno_type(value):
    """Creates a UNO type from gdb.Value of type
        com::sun::star::uno::Type, typelib_TypeDescription, or
        typelib_TypeDescriptionReference
    """

    # Extraction of value of type struct _typelib_TypeDescription
    type = value.type.strip_typedefs()

    cssu_type = 'com::sun::star::uno::Type'
    type_desc = '_typelib_TypeDescription'
    type_descs = (
            type_desc,
            '_typelib_CompoundTypeDescription',
            '_typelib_StructTypeDescription',
            '_typelib_IndirectTypeDescription',
            '_typelib_EnumTypeDescription',
            '_typelib_InterfaceMemberTypeDescription',
            '_typelib_InterfaceMethodTypeDescription',
            '_typelib_InterfaceAttributeTypeDescription',
            '_typelib_InterfaceTypeDescription'
    )
    type_desc_ref = '_typelib_TypeDescriptionReference'

    if type.tag == cssu_type:
        type_ref = type['_pType']
    while type.tag == type_desc_ref:
        pvalue = value['pType']
        assert pvalue
        value = pvalue.dereference()
        type = value.type.strip_typedefs()
    if str(type.tag) not in type_descs:
        return None

    # Determination of the UNO type

    full_value = value
    if type.tag != type_desc:
        while 'aBase' in value:
            value = value['aBase']
    type_class = int(value['eTypeClass'])
    name = make_string(value['pTypeName'])
    uno_type = None
    if type_class == TypeCode.VOID:
        uno_type = VoidType()
    elif type_class == TypeCode.CHAR:
        uno_type = PrimitiveType(type_class, name, 'sal_Char')
    elif type_class == TypeCode.BOOLEAN:
        uno_type = PrimitiveType(type_class, name, 'sal_Bool')
    elif type_class == TypeCode.BYTE:
        uno_type = PrimitiveType(type_class, name, 'sal_Int8')
    elif type_class == TypeCode.SHORT:
        uno_type = PrimitiveType(type_class, name, 'sal_Int16')
    elif type_class == TypeCode.UNSIGNED_SHORT:
        uno_type = PrimitiveType(type_class, name, 'sal_uInt16')
    elif type_class == TypeCode.LONG:
        uno_type = PrimitiveType(type_class, name, 'sal_Int32')
    elif type_class == TypeCode.UNSIGNED_LONG:
        uno_type = PrimitiveType(type_class, name, 'sal_uInt32')
    elif type_class == TypeCode.HYPER:
        uno_type = PrimitiveType(type_class, name, 'sal_Int64')
    elif type_class == TypeCode.UNSIGNED_HYPER:
        uno_type = PrimitiveType(type_class, name, 'sal_uInt64')
    elif type_class == TypeCode.FLOAT:
        uno_type = PrimitiveType(type_class, name, 'float')
    elif type_class == TypeCode.DOUBLE:
        uno_type = PrimitiveType(type_class, name, 'double')
    elif type_class == TypeCode.STRING:
        uno_type = PrimitiveType(type_class, name, 'rtl::OUString')
    elif type_class == TypeCode.TYPE:
        uno_type = PrimitiveType(type_class, name, 'com::sun::star::uno::Type')
    elif type_class == TypeCode.ANY:
        uno_type = PrimitiveType(type_class, name, 'com::sun::star::uno::Any')
    elif type_class == TypeCode.ENUM:
        uno_type = EnumType(value, full_value)
    elif type_class == TypeCode.TYPEDEF:
        pass
    elif type_class == TypeCode.STRUCT:
        uno_type = StructType(value, full_value)
    elif type_class == TypeCode.UNION:
        raise UnsupportedType('union')
    elif type_class == TypeCode.EXCEPTION:
        uno_type = CompoundType(value, full_value)
    elif type_class == TypeCode.SEQUENCE:
        uno_type = IndirectType(value, full_value)
    elif type_class == TypeCode.ARRAY:
        raise UnsupportedType('array')
    elif type_class == TypeCode.INTERFACE:
        uno_type = InterfaceType(value, full_value)
    elif type_class == TypeCode.SERVICE:
        raise UnsupportedType('service')
    elif type_class == TypeCode.MODULE:
        raise UnsupportedType('module')
    elif type_class == TypeCode.INTERFACE_METHOD:
        uno_type = InterfaceMethodType(value, full_value)
    elif type_class == TypeCode.INTERFACE_ATTRIBUTE:
        uno_type = InterfaceAttributeType(value, full_value)
    elif type_class == TypeCode.UNKNOWN:
        raise UnknownType(type)
    elif type_class == TypeCode.PROPERTY:
        pass
    elif type_class == TypeCode.CONSTANT:
        pass
    elif type_class == TypeCode.CONSTANTS:
        pass
    elif type_class == TypeCode.SINGLETON:
        pass
    else:
        raise UnknownType(type)

    assert uno_type
    return uno_type


def uno_cast(type, value):
    """Casts value or pointer to UNO type represented by type."""
    if value.type.code == gdb.TYPE_CODE_PTR:
        return value.cast(type.to_gdb().pointer())
    else:
        return value.cast(type.to_gdb())


class VoidType (Type):

    def __init__(self):
        Type.__init__(self, TypeCode.VOID, "void")
        self.typename = "void"


class PrimitiveType (Type):

    def __init__(self, code, typename_uno, typename_cpp):
        Type.__init__(self, code, typename_uno)
        self.typename = str(typename_cpp)


class CompoundType (Type):

    def __init__(self, type, full_type):
        Type.__init__(self, int(type['eTypeClass']),
                make_string(type['pTypeName'].dereference()))
        self.typename = self.tag.replace('.', '::')
        self._type = full_type

    class _iterator:

        def __init__(self, count, types, names):
            self.count = count
            self.members = members
            self.names = names
            self.pos = 0

        def __iter__(self):
            return self

        def next(self):
            assert self.pos >= 0 and self.pos <= self.count
            if self.pos == self.count:
                raise StopIteration

            pmember = self.members[self.pos]
            assert pmember
            pname = self.names[self.i]
            assert pname
            self.pos = self.pos + 1
            member = make_uno_type(pmember.dereference())
            assert member
            name = make_string(pname.dereference())
            return (name, member)


    def attributes(self):
        return _iterator(self._type['nMembers'], self._type['ppTypeRefs'],
                self._type['ppMemberNames'])


class StructType (CompoundType):

    def __init__(self, type, full_type):
        CompoundType.__init__(self, type, full_type['aBase'])


class IndirectType (Type):

    def __init__(self, type, full_type):
        Type.__init__(self, type['eTypeClass'],
                make_string(type['pTypeName'].dereference()))
        pelem = self.full_type['pType']
        assert pelem
        self.element = make_uno_type(pelem.dereference())
        assert self.element
        self.typename = ('com::sun::star::uno::Sequence<'
                + self.element.typename + '>')


class EnumType (Type):

    def __init__(self, type, full_type):
        Type.__init__(self, TypeCode.ENUM,
                make_string(type['pTypeName'].dereference()))
        self.typename = self.tag.replace('.', '::')
        self._type = full_type

    class _iterator:

        def __init__(self, count, values, names):
            self.count = count
            self.values = values
            self.names = names
            self.pos = 0

        def __iter__(self):
            return self

        def next(self):
            assert self.pos >= 0 and self.pos <= self.count
            if self.pos == self.count:
                raise StopIteration

            pvalue = self.values[self.pos]
            assert pvalue
            pname = self.names[self.pos]
            assert pname
            self.pos = self.pos + 1
            value = int(pvalue.dereference())
            name = make_string(pname.dereference())
            return (name, value)

    def values(self):
        return _iterator(self._type['nEnumValues'],
                self._type['ppEnumNames'], self._type['pEnumValues'])

    def default_value(self):
        return self._type['nDefaultEnumValue']


class InterfaceMemberType (Type):

    def __init__(self, type, full_type):
        Type.__init__(self, type['eTypeClass'],
                make_string(type['pTypeName'].dereference()))
        (interface, delim, member) = self.tag.partition('::')
        self.typename = interface.replace('.', '::') + '::*' + member
        self.position = type['nPosition']
        pname = type['pMemberName']
        assert pname
        self.name = make_string(pname.dereference())


class InterfaceMethodType (InterfaceMemberType):

    def __init__(self, type, full_type):
        InterfaceMemberType.__init__(self, type, full_type['aBase'])
        pret = full_type['pReturnTypeRef']
        assert pret
        self.return_type = make_uno_type(pret.dereference())
        assert self.return_type
        self.oneway = bool(full_type['bOneWay'])
        self._type = full_type

    class _iterator:

        def __init__(self, count, values):
            self.count = count
            self.values = values
            self.pos = 0
            assert values

        def __iter__(self):
            return self

        def next(self):
            assert self.pos >= 0 and self.pos <= self.count
            if self.pos == self.count:
                raise StopIteration

            value = self.values[self.pos]
            self.pos = self.pos + 1
            return value


    class parameter (tuple):

        def __init__(self, type):
            self.__init_tuple(type)
            self.input = type['bIn']
            self.output = type['bOut']

        def _init_tuple(self, type):
            pname = self['pName']
            assert pname
            ptype = self['pTypeRef']
            assert ptype
            name = make_string(pname.dereference())
            type = make_uno_type(ptype.dereference())
            assert type
            tuple.__init__(self, name, type)


    def parameters(self):
        for param in _iterator(self._type['nParams'], self._type['pParams']):
            yield parameter(param)

    def exceptions(self):
        def make_exception(self, pex):
            assert pex
            ex = make_uno_type(pex.dereference())
            assert ex
            return ex

        for ex in _iterator(
                self._type['nExceptions'], self._type['ppExceptions']):
            yield make_exception(ex)


class InterfaceAttributeType (InterfaceMemberType):

    def __init__(self, type, full_type):
        InterfaceMemberType.__init__(self, type, full_type['aBase'])
        self.readonly = full_type['bReadOnly']
        ptype = full_type['pAttributeTypeRef']
        assert ptype
        self.type = make_uno_type(ptype.dereference())
        assert self.type


class MembersNotInitialized (Exception):
    """Represents exception raised when interface type' members haven't
        been initialized (i.e. just level 1 initialization has been
        performed)."""
    pass

class InterfaceType (Type):

    def __init__(self, type, full_type):
        
        Type.__init__(self, TypeCode.INTERFACE,
                make_string(type['pTypeName'].dereference()))
        assert int(type['eTypeClass']) == TypeCode.INTERFACE
        self.typename = self.tag.replace('.', '::')
        self.uik = full_type['aUik']
        self._type = full_type

    class _iterator:

        def __init__(self, count, values):
            assert values
            self.count = count
            self.values = values
            self.pos = 0

        def __iter__(self):
            return self

        def next(self):
            assert self.pos >= 0 and self.pos <= self.count
            pvalue = self.values[self.pos]
            assert pvalue
            self.pos = self.pos + 1
            uno = make_uno_type(pvalue.dereference())
            assert uno
            return uno


    def members(self):
        return __members(self._type['nMembers'], self._type['ppMembers'])

    def all_members(self):
        return __members(self._type['nAllMembers'], self._type['ppAllMembers'])

    def __members(count, values):
        if values == 0:
            raise MembersNotInitialized
        return _iterator(count, values)

    def bases(self):
        return _iterator(self._type['nBaseTypes'], self._type['ppBaseTypes'])
